from typing import List, TypeVar, Union

from ..base import *
from ..models import *
from .mixins import *

T = TypeVar("T")


class Workflow(ParentMixin, SaveModelMixin, DeleteModelMixin, BaseModel):
    def __init__(
        self,
        url=None,
        id=None,
        name=None,
        title=None,
        tasks=None,
        error_task=None,
        task_defaults=None,
        variables=None,
        usages=None,
        project=None,
        project_id=None,
        created_at=None,
        updated_at=None,
        last_execution=None,
        description=None,
        extra=None,
        labels=None,
        permissions=None,
        version=None,
        updated_by=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.url = url
        self.id = id
        self.name = name
        self.title = title
        self.tasks = tasks
        self.error_task = error_task
        self.task_defaults = task_defaults
        self.variables = variables
        self.usages = usages
        self.project = project
        self.project_id = project_id
        self.created_at = created_at
        self.updated_at = updated_at
        self.last_execution = last_execution
        self.description = description
        self.extra = extra
        self.labels = labels
        self.permissions = permissions
        self.version = version
        self.updated_by = updated_by

    _read_only = [
        "url",
        "id",
        "usages",
        "project",
        "project_id",
        "created_at",
        "updated_by",
        "updated_at",
        "last_execution",
        "permissions",
        "version",
    ]

    _submodels = [Task]

    def _tasks(self, identifier=None, **kwargs) -> Union[Task, List[Task]]:
        return self._get(Task, identifier, **kwargs)

    @operation
    def run(self, input=None) -> Execution:
        pass

    @operation
    def duplicate(self: T, name=None, title=None) -> T:
        pass



from ..base import *
from ..models import *
from .mixins import *


class WebhookLogResource(
    GetResourceMixin, BaseResource,
):
    def __init__(self, parent: BaseResource):
        super().__init__("log", WebhookLog, parent.client)
        self.parent = parent


Mapping.register(WebhookLog, WebhookLogResource)



from ..base import *
from ..models import *
from ..njinn_client import NjinnClient
from .mixins import *


class WorkflowResource(
    GetResourceMixin,
    SaveResourceMixin,
    CreateResourceMixin,
    DeleteResouceMixin,
    BaseResource,
):
    def __init__(self, client: NjinnClient):
        super().__init__("api/v1/workflows", Workflow, client)

    def run(self, **kwargs):
        return self.operation("run", Execution).post(**kwargs)

    def duplicate(self, **kwargs):
        return self.operation("duplicate", Workflow).post(**kwargs)


Mapping.register(Workflow, WorkflowResource)



from ..base import *
from ..models import *
from ..njinn_client import NjinnClient
from .mixins import *


class WebhookResource(
    GetResourceMixin,
    SaveResourceMixin,
    CreateResourceMixin,
    DeleteResouceMixin,
    BaseResource,
):
    def __init__(self, client: NjinnClient):
        super().__init__("api/v1/hooks", Webhook, client)

    def duplicate(self, **kwargs) -> Webhook:
        return self.operation("duplicate", Webhook).post(**kwargs)


Mapping.register(Webhook, WebhookResource)



from ..base import *
from ..models import *
from .mixins import *


class Execution(ParentMixin, SaveModelMixin, BaseModel):
    def __init__(
        self,
        url=None,
        id=None,
        workflow=None,
        labels=None,
        state=None,
        state_info=None,
        input=None,
        params=None,
        started_at=None,
        ended_at=None,
        runtime=None,
        trigger=None,
        webhook=None,
        schedule=None,
        task=None,
        parent_task_execution=None,
        parent_execution=None,
        result=None,
        user=None,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.url = url
        self.id = id
        self.workflow = workflow
        self.labels = labels
        self.state = state
        self.state_info = state_info
        self.input = input
        self.params = params
        self.started_at = started_at
        self.ended_at = ended_at
        self.runtime = runtime
        self.trigger = trigger
        self.webhook = webhook
        self.schedule = schedule
        self.task = task
        self.parent_task_execution = parent_task_execution
        self.parent_execution = parent_execution
        self.result = result
        self.user = user

    _read_only = [
        "url",
        "id",
        "runtime",
        "task",
        "parent_task_execution",
        "parent_execution",
    ]

    _submodels = [TaskExecution]

    def tasks(self, identifier=None, **kwargs) -> Union[TaskExecution, List[TaskExecution]]:
        return self._get(TaskExecution, identifier, **kwargs)

    @operation
    def cancel(self) -> None:
        pass

    @operation
    def pause(self) -> None:
        pass

    @operation
    def resume(self) -> None:
        pass

    @operation
    def log(self) -> str:
        pass



from typing import List, Union

from .base import *
from .models import *
from .models.mixins import ParentMixin
from .njinn_client import NjinnClient
from .resources import *


class NjinnAPI(ParentMixin):
    _submodels = [Workflow, Config, Execution, NjinnGroup, Webhook, Schedule, Rule]

    def __init__(
        self, host: str, username: str = None, password: str = None, token: str = None
    ):
        self.client = NjinnClient(
            host, username=username, password=password, token=token
        )
        self.api_resource = RootResource(self.client)

    def workflows(self, identifier=None, **kwargs) -> Union[Workflow, List[Workflow]]:
        return self._get(Workflow, identifier, **kwargs)

    def configs(self, identifier=None, **kwargs) -> Union[Config, List[Config]]:
        return self._get(Config, identifier, **kwargs)

    def executions(
        self, identifier=None, **kwargs
    ) -> Union[Execution, List[Execution]]:
        return self._get(Execution, identifier, **kwargs)

    def groups(self, identifier=None, **kwargs) -> Union[NjinnGroup, List[NjinnGroup]]:
        return self._get(NjinnGroup, identifier, **kwargs)

    def hooks(self, identifier=None, **kwargs) -> Union[Webhook, List[Webhook]]:
        return self._get(Webhook, identifier, **kwargs)

    def schedules(self, identifier=None, **kwargs) -> Union[Schedule, List[Schedule]]:
        return self._get(Schedule, identifier, **kwargs)

    def scheduling_rules(self, identifier=None, **kwargs) -> Union[Rule, List[Rule]]:
        return self._get(Rule, identifier, **kwargs)

    @operation
    def schedules_preview(self, schedule=None) -> dict:
        pass

    @operation
    def scheduling_rules_preview(
        self,
        name=None,
        title=None,
        rule_type=None,
        rrule=None,
        grouping_rule=None,
        fixed_offset_rule=None,
        relative_offset_rule=None,
        project=None,
        business_calendar=None,
        labels=None,
        description=None,
        count=None,
    ) -> dict:
        pass



from typing import TypeVar

from ..base import *
from ..models import *
from .mixins import *

T = TypeVar("T")


class Webhook(ParentMixin, SaveModelMixin, DeleteModelMixin, BaseModel):
    def __init__(
        self,
        id=None,
        version=None,
        updated_by=None,
        last_call=None,
        created_at=None,
        updated_at=None,
        labels=None,
        name=None,
        title=None,
        is_active=None,
        key=None,
        source_ip_filter=None,
        header_validation_name=None,
        header_validation_value=None,
        inputs=None,
        description=None,
        workflow=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.id = id
        self.version = version
        self.updated_by = updated_by
        self.last_call = last_call
        self.created_at = created_at
        self.updated_at = updated_at
        self.labels = labels
        self.name = name
        self.title = title
        self.is_active = is_active
        self.key = key
        self.source_ip_filter = source_ip_filter
        self.header_validation_name = header_validation_name
        self.header_validation_value = header_validation_value
        self.inputs = inputs
        self.description = description
        self.workflow = workflow

    _read_only = [
        "id",
        "version",
        "created_at",
        "updated_by",
        "updated_at",
        "last_call",
    ]

    _submodels = [WebhookLog]

    def logs(self, identifier=None, **kwargs) -> Union[WebhookLog, List[WebhookLog]]:
        return self._get(WebhookLog, identifier, **kwargs)

    @operation
    def duplicate(self: T, name=None, title=None) -> T:
        pass




from ..base import *
from .mixins import *


class TaskExecution(BaseModel):
    def __init__(
        self,
        id=None,
        name=None,
        execution=None,
        state=None,
        state_info=None,
        input=None,
        started_at=None,
        ended_at=None,
        runtime=None,
        triggered_by=None,
        position=None,
        active=None,
        result=None,
        queue=None,
        published=None,
        is_error_task=None,
        on_success=None,
        on_complete=None,
        on_error=None,
        action=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.id = id
        self.execution = execution
        self.name = name
        self.state = state
        self.state_info = state_info
        self.input = input
        self.started_at = started_at
        self.ended_at = ended_at
        self.runtime = runtime
        self.triggered_by = triggered_by
        self.position = position
        self.active = active
        self.result = result
        self.queue = queue
        self.published = published
        self.is_error_task = is_error_task
        self.action = action
        self.on_success = on_success
        self.on_complete = on_complete
        self.on_error = on_error

    _property_map = {
        "on_success": "on-success",
        "on_complete": "on-complete",
        "on_error": "on-error",
    }

    _read_only = [
        "id",
        "started_at",
        "ended_at",
        "runtime",
        "position",
        "active",
        "queue",
        "published",
        "on-complete",
        "on-error",
        "action",
    ]

    @operation
    def log() -> str:
        pass

    @operation
    def cancel() -> None:
        pass




from ..base import *
from ..models import *
from ..njinn_client import NjinnClient
from .mixins import *


class NjinnGroupResource(
    GetResourceMixin,
    SaveResourceMixin,
    CreateResourceMixin,
    DeleteResouceMixin,
    BaseResource,
):
    def __init__(self, client: NjinnClient):
        super().__init__("api/v1/groups", NjinnGroup, client)

Mapping.register(NjinnGroup, NjinnGroupResource)



from ..base import *
from ..models import *
from ..njinn_client import NjinnClient
from .mixins import *


class RootResource:
    def __init__(self, client: NjinnClient):
        self.client = client

    def operation(self, path: str, result_class: Type = None):
        return Operation(self.client, path, result_class)

    def schedules_preview(self, **kwargs):
        return self.operation("api/v1/schedules/preview").post(**kwargs)

    def scheduling_rules_preview(self, **kwargs):
        return self.operation("api/v1/scheduling_rules/preview").post(**kwargs)




from .config import Config
from .task_execution import TaskExecution
from .execution import Execution
from .group_member import GroupMember
from .njinn_group import NjinnGroup
from .task import Task
from .workflow import Workflow
from .webhooklog import WebhookLog
from .webhook import Webhook
from .schedule import Schedule
from .rule import Rule


from ..base import *
from ..models import *
from ..njinn_client import NjinnClient
from .mixins import *


class ConfigResource(
    GetResourceMixin,
    SaveResourceMixin,
    CreateResourceMixin,
    DeleteResouceMixin,
    BaseResource,
):
    def __init__(self, client: NjinnClient):
        super().__init__("api/v1/configs", Config, client)

    def duplicate(self, **kwargs) -> Config:
        return self.operation("duplicate", Config).post(**kwargs)


Mapping.register(Config, ConfigResource)



from .root_resource import RootResource
from .config_resource import ConfigResource
from .task_execution_resource import TaskExecutionResource
from .execution_resource import ExecutionResource
from .group_member_resource import GroupMemberResource
from .njinn_group_resource import NjinnGroupResource
from .task_resource import TaskResource
from .workflow_resource import WorkflowResource
from .webhook_resource import WebhookResource
from .webhooklog_resource import WebhookLogResource
from .schedule_resource import ScheduleResource
from .rule_resource import RuleResource



from typing import TypeVar

from ..base import *
from ..models import *
from .mixins import *

T = TypeVar("T")


class WebhookLog(BaseModel):
    def __init__(
        self,
        id=None,
        created_at=None,
        updated_at=None,
        source_ip=None,
        http_request_method=None,
        data=None,
        log=None,
        webhook=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.id = id
        self.created_at = created_at
        self.updated_at = updated_at
        self.source_ip = source_ip
        self.http_request_method = http_request_method
        self.data = data
        self.log = log
        self.webhook = webhook



import json
import logging
from enum import Enum

import requests

logger = logging.getLogger("njinn")


class ContentType(Enum):
    JSON = (1,)
    TEXT = (2,)


class NjinnClient:
    def __init__(
        self, host: str, username: str = None, password: str = None, token: str = None
    ) -> None:
        self.host = host.strip()
        if self.host.endswith("/"):
            self.host = self.host[:-1]

        self._headers = {"Content-Type": "application/json"}

        if token is None:
            assert username is not None and password is not None
            self._auth = (username, password)
            self._headers = None
        else:
            self._auth = None
            self._headers["Authorization"] = f"JWT {token}"

    def __get_full_url(self, url: str):
        return f"{self.host}/{url}"

    def get(self, url, **kwargs):
        return self.request("GET", url, **kwargs)

    def post(self, url, data, **kwargs):
        return self.request("POST", url, json=data, **kwargs)

    def put(self, url, data, **kwargs):
        return self.request("PUT", url, json=data, **kwargs)

    def patch(self, url, data, **kwargs):
        return self.request("PATCH", url, json=data, **kwargs)

    def delete(self, url, **kwargs):
        return self.request("DELETE", url, **kwargs)

    def request(
        self, method, url, no_auth=False, response_content_type=ContentType.JSON, **kwargs
    ):
        headers = self._headers
        if not no_auth:
            if self._auth is not None:
                kwargs.setdefault("auth", self._auth)
        else:
            del headers["Authorization"]

        logger.debug(f"Sending request: {method} {url}")
        logger.debug(json.dumps(kwargs.get("json"), indent=2))

        response = requests.request(
            method, self.__get_full_url(url), headers=headers, **kwargs
        )

        try:
            response.raise_for_status()
        except Exception:
            logger.error(response.text)
            raise

        if len(response.content) > 0:
            if response_content_type == ContentType.JSON:
                try:
                    return response.json()
                except Exception:
                    raise Exception("Error parsing response: " + response.text)
            else:
                return response.text
        else:
            return None



from ..base import *
from .mixins import *


class GroupMember(SaveModelMixin, DeleteModelMixin, BaseModel):
    def __init__(self, user=None, group=None, **kwargs):
        super().__init__(**kwargs)
        self.user = user
        self.group = group

    @property
    def api_identifier(self):
        return self.user



from typing import TypeVar

from ..base import *
from .mixins import *

T = TypeVar("T")


class Rule(SaveModelMixin, DeleteModelMixin, BaseModel):
    def __init__(
        self,
        id=None,
        name=None,
        created_at=None,
        updated_at=None,
        updated_by=None,
        version=None,
        title=None,
        rule_type=None,
        rrule=None,
        grouping_rule=None,
        fixed_offset_rule=None,
        relative_offset_rule=None,
        project=None,
        business_calendar=None,
        labels=None,
        description=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.id = id
        self.created_at = created_at
        self.updated_at = updated_at
        self.updated_by = updated_by
        self.name = name
        self.version = version
        self.title = title
        self.rule_type = rule_type
        self.rrule = rrule
        self.grouping_rule = grouping_rule
        self.fixed_offset_rule = fixed_offset_rule
        self.relative_offset_rule = relative_offset_rule
        self.project = project
        self.business_calendar = business_calendar
        self.labels = labels
        self.description = description

    _read_only = [
        "id",
        "version",
        "updated_by",
        "updated_at",
        "created_at",
    ]

    @operation
    def duplicate(self: T, name=None, title=None) -> T:
        pass



from typing import TypeVar

from ..base import *
from .mixins import *

T = TypeVar("T")


class Config(SaveModelMixin, DeleteModelMixin, BaseModel):

    def __init__(
        self,
        url=None,
        id=None,
        project=None,
        name=None,
        title=None,
        values=None,
        config_scheme=None,
        labels=None,
        description=None,
        version=None,
        updated_by=None,
        updated_at=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.url = url
        self.id = id
        self.project = project
        self.name = name
        self.title = title
        self.values = values
        self.config_scheme = config_scheme
        self.labels = labels
        self.description = description
        self.version = version
        self.updated_by = updated_by
        self.updated_at = updated_at

    _read_only = [
        "url",
        "id",
        "version",
        "updated_by",
        "updated_at",
        "created_at",
    ]
    
    @operation
    def duplicate(self: T, name=None, title=None) -> T:
        pass



import functools
import inspect
from copy import deepcopy
from typing import List, Type, TypeVar, Union

T = TypeVar("T")

from .njinn_client import ContentType, NjinnClient


class BaseResource:
    def __init__(self, path: str, model_class: Type, client: NjinnClient) -> None:
        self._path = path
        self.model_class = model_class
        self.client = client
        self.identifier = None

    @property
    def path(self) -> str:
        def with_identifier():
            return f"/{self.identifier}" if self.identifier else ""

        parent = getattr(self, "parent", None)
        if parent:
            return f"{parent.path}/{self._path}{with_identifier()}"
        return f"{self._path}{with_identifier()}"

    def construct(self, response):
        return self.model_class(api_resource=self, **response)

    def operation(self, path: str, result_class: Type = None):
        return Operation(self.client, f"{self.path}/{path}", result_class)


def model_init(func):
    """
    Swap kwargs based on _property_map to avoid aguments containing special characters like hyphen
    """

    @functools.wraps(func)
    def __init__(self, *args, **kwargs):
        for prop, attr in self._property_map.items():
            if attr in kwargs:
                kwargs.setdefault(prop, kwargs.pop(attr))

        # Call original __init__
        func(self, *args, **kwargs)

    return __init__


class BaseModel:
    _internal = ["_api_resource", "_read_only", "_submodels", "_property_map"]
    _read_only = []
    _submodels = []
    _property_map = {}

    def __new__(cls, *args, **kwargs):
        """
        Declare properties and backing attributes based on _property_map
        """
        for prop, attr in cls._property_map.items():
            setattr(cls, attr, None)
            setattr(
                cls,
                prop,
                property(
                    fget=lambda obj, key=attr: getattr(obj, key),
                    fset=lambda obj, value, key=attr: setattr(obj, key, value),
                ),
            )

        # Decorate __init__ to treat special arguments
        cls.__init__ = model_init(cls.__init__)
        return object.__new__(cls)

    def __init__(self, api_resource: BaseResource = None) -> None:
        self._api_resource = api_resource

    @property
    def api_identifier(self):
        return getattr(self, "id", None)

    @property
    def api_resource(self) -> BaseResource:
        if self._api_resource is None:
            if self.api_identifier is None:
                raise Exception(
                    f"{self} does not have '_api_resource'. Object needs to be created by NjinnAPI or parent resource first."
                )
            raise Exception(f"{self} does not have '_api_resource'")

        def renew_api_resource():
            if self._api_resource.identifier != self.api_identifier:
                self._api_resource = deepcopy(self._api_resource)
                self._api_resource.identifier = self.api_identifier
            return self._api_resource

        return renew_api_resource()

    def refresh(self: T) -> T:
        return self._refresh(self.api_resource.get(self.api_identifier))

    def _refresh(self: T, instance: T) -> T:
        for attr in self.__dict__:
            if hasattr(instance, attr):
                self.__setattr__(attr, deepcopy(getattr(instance, attr)))

        return self


class Mapping:
    _resource_map = {}

    @staticmethod
    def register(model_class: Type, resource_class):
        Mapping._resource_map[model_class] = resource_class

    @staticmethod
    def lookup_resource(model_class: Type) -> Type:
        if model_class in Mapping._resource_map.keys():
            return Mapping._resource_map[model_class]

        raise NotImplementedError(f"{model_class} has no backing resource")


class ResourceUtil:
    def __init__(
        self,
        resource_class: Type[BaseResource],
        context: Union[NjinnClient, BaseResource],
    ):
        self.resource_class = resource_class
        self.context = context

    def get(self, identifier=None, **kwargs) -> Union[BaseModel, List[BaseModel]]:

        resource = self.resource_class(self.context)
        return resource.get(identifier, **kwargs)

    def create(self, obj: BaseModel) -> BaseModel:
        resource = self.resource_class(self.context)
        return resource.create(obj)

    @classmethod
    def to_json(cls, obj: dict, exclude_none=False, fields: List = None):
        copy = obj.copy() if type(obj) is dict else obj.__dict__.copy()

        is_base_model = issubclass(type(obj), BaseModel)
        excluded_fields = (obj._read_only + obj._internal) if is_base_model else []

        json = {}
        for key in copy:
            if fields and key not in fields:
                continue

            if key in excluded_fields:
                continue

            value = copy[key]
            if exclude_none and value is None:
                continue

            if type(value) is dict or hasattr(value, "__dict__"):
                json[key] = ResourceUtil.to_json(value, exclude_none=exclude_none)
            else:
                json[key] = value

        return json


class Operation:
    def __init__(
        self, client: NjinnClient, path: str, result_class: Type = None,
    ):
        self.client = client
        self.path = path
        self.result_class = result_class

    def post(self, **kwargs):
        body = ResourceUtil.to_json(kwargs, exclude_none=True)
        response = self.client.post(self.path, body)
        return self.construct(response)

    def get(self, **kwargs):
        if self.result_class is str:
            response = self.client.get(
                self.path, response_content_type=ContentType.TEXT, params=kwargs
            )
        else:
            response = self.client.get(self.path, params=kwargs)
        return self.construct(response)

    def construct(self, response):
        if self.result_class is None or self.result_class is str:
            return response

        if issubclass(self.result_class, BaseModel):
            resource_class = Mapping.lookup_resource(self.result_class)
            return resource_class(self.client).construct(response)

        return self.result_class(**response)


def operation(func):
    @functools.wraps(func)
    def resource_operation(*args, **kwargs):
        args_name = list(inspect.signature(func).parameters.keys())
        args_dict = dict(zip(args_name[1:], args[1:]))
        kwargs.update(args_dict)

        operation = getattr(args[0].api_resource, func.__name__)
        return operation(**kwargs)

    return resource_operation



from ..base import *
from ..models import *
from .mixins import *


class GroupMemberResource(
    GetResourceMixin,
    SaveResourceMixin,
    CreateResourceMixin,
    DeleteResouceMixin,
    BaseResource,
):
    def __init__(self, parent: BaseResource):
        super().__init__("members", GroupMember, parent.client)
        self.parent = parent

Mapping.register(GroupMember, GroupMemberResource)



from typing import TypeVar

from ..base import *
from .mixins import *

T = TypeVar("T")


class Schedule(SaveModelMixin, DeleteModelMixin, BaseModel):
    def __init__(
        self,
        id=None,
        name=None,
        created_at=None,
        updated_at=None,
        is_active=None,
        schedule_type=None,
        cron=None,
        time=None,
        interval_minutes=None,
        between_start=None,
        between_end=None,
        rule=None,
        filter_parameters=None,
        timezone=None,
        inputs=None,
        workflow=None,
        title=None,
        labels=None,
        next_execution=None,
        last_execution=None,
        version=None,
        updated_by=None,
        description=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.id = id
        self.name = name
        self.created_at = created_at
        self.updated_at = updated_at
        self.is_active = is_active
        self.schedule_type = schedule_type
        self.cron = cron
        self.time = time
        self.interval_minutes = interval_minutes
        self.between_start = between_start
        self.between_end = between_end
        self.rule = rule
        self.filter_parameters = filter_parameters
        self.timezone = timezone
        self.inputs = inputs
        self.workflow = workflow
        self.title = title
        self.labels = labels
        self.next_execution = next_execution
        self.last_execution = last_execution
        self.version = version
        self.updated_by = updated_by
        self.description = description

    _read_only = [
        "id",
        "version",
        "updated_by",
        "updated_at",
        "created_at",
        "last_execution",
    ]

    @operation
    def duplicate(self: T, name=None, title=None) -> T:
        pass



from ..base import *
from ..models import *
from ..njinn_client import NjinnClient
from .mixins import *


class ExecutionResource(GetResourceMixin, SaveResourceMixin, BaseResource):
    def __init__(self, client: NjinnClient):
        super().__init__("api/v1/executions", Execution, client)

    def cancel(self, **kwargs) -> None:
        return self.operation("cancel").post(**kwargs)

    def pause(self, **kwargs) -> None:
        return self.operation("pause").post(**kwargs)

    def resume(self, **kwargs) -> None:
        return self.operation("resume").post(**kwargs)
    
    def log(self, **kwargs) -> str:
        return self.operation("log", result_class=str).get(**kwargs)


Mapping.register(Execution, ExecutionResource)



from typing import List

from ..base import *


class GetResourceMixin(BaseResource):
    def get(self, identifier=None, **kwargs):
        if identifier is None:
            response = self.client.get(self.path, params=kwargs)
            if response.get("results"):
                return [self.construct(entity) for entity in response.get("results")]
            return [self.construct(response[key]) for key in response]
        else:
            self.identifier = identifier
            response = self.client.get(f"{self.path}")
            return self.construct(response)


class SaveResourceMixin(BaseResource):
    def save(self, obj: BaseModel, fields: List = None):
        if fields:
            body = ResourceUtil.to_json(obj, fields=fields)
            response = self.client.patch(f"{self.path}", body)
        else:
            body = ResourceUtil.to_json(obj)
            response = self.client.put(f"{self.path}", body)

        return self.construct(response)


class CreateResourceMixin(BaseResource):
    def create(self, obj: BaseModel):
        body = ResourceUtil.to_json(obj, exclude_none=True)
        response = self.client.post(f"{self.path}", body)
        return self.construct(response)


class DeleteResouceMixin(BaseResource):
    def delete(self, obj: BaseModel):
        return self.client.delete(f"{self.path}")



from typing import List, Type, TypeVar

from ..base import *

T = TypeVar("T")


class ParentMixin:
    @property
    def _context(self):
        if issubclass(type(self), BaseModel):
            return self.api_resource
        return self.client

    def create(self, obj: T) -> T:
        self._guard_creation(type(obj))
        return obj._refresh(self._init_subresource(type(obj)).create(obj))

    def _get(self, model_class: Type, identifier=None, **kwargs) -> T:
        return self._init_subresource(model_class).get(identifier, **kwargs)

    def _init_subresource(self, model_class: Type):
        return ResourceUtil(Mapping.lookup_resource(model_class), self._context)

    def _guard_creation(self, model_class: Type):
        types_with_create = [
            type
            for type in self._submodels
            if hasattr(Mapping.lookup_resource(type), "create")
        ]

        if model_class not in types_with_create:
            message = f"{model_class.__name__} cannot be created from {self.__class__.__name__}."
            message += f" Valid options: {', '.join([type.__name__ for type in types_with_create])}"
            raise NotImplementedError(message)


class SaveModelMixin(BaseModel):
    def save(self: T, fields: List = None) -> T:
        return self._refresh(self.api_resource.save(self, fields))


class DeleteModelMixin(BaseModel):
    def delete(self):
        return self.api_resource.delete(self)



from collections import UserDict

from ..base import *


class Task(BaseModel, UserDict):
    def __init__(self, dict, **kwargs) -> None:
        super().__init__(**kwargs)
        self.data = {}
        if dict is not None:
            self.update(dict)

    @property
    def api_identifier(self):
        return self.get("name", None)



import getpass
import logging
import os
import shutil
import stat
import tempfile
from urllib.parse import urlparse
from zipfile import ZipFile

import click
import git
import requests
import yaml

logger = logging.getLogger(__name__)

NJINN_NAMESPACE = "njinn"


def login(username, password, base_api):
    auth_url = base_api + "/api/token/auth"

    credentials = {"username": username, "password": password}

    auth_response = requests.post(auth_url, json=credentials)

    if not auth_response.ok:
        raise click.ClickException("Wrong credentials")

    jwt_token = auth_response.json()["token"]

    headers = {"Authorization": f"JWT {jwt_token}"}

    global r
    r = requests.Session()
    r.headers.update(headers)


def prepare_actions(config):
    actions_data = []
    actions = config["actions"]

    for (key, value) in actions.items():
        action = {
            "name": key,
            "title": value.get("title"),
            "description": value.get("description", ""),
            "icon": value.get("icon", ""),
            "path": value.get("path"),
            "parameters": value.get("parameters", {}),
            "output": value.get("output", {}),
            "platform": value.get("platform"),
        }
        actions_data.append(action)

    logger.info(f"Obtained actions: {len(actions_data)}")

    return actions_data


def prepare_config_schemes(config):
    config_schemes_data = []
    config_schemes = config.get("config-schemes", {})

    for key, value in config_schemes.items():
        config_scheme = {
            "name": key,
            "input_scheme": value,
            "title": value.get("title", key),
            "description": value.get("description", ""),
            "icon": value.get("icon", ""),
        }
        config_schemes_data.append(config_scheme)

    logger.info(f"Obtained config schemes: {len(config_schemes_data)}")

    return config_schemes_data


def prepare_pack(config, repository, branch, version_identifier):
    pack_data = {
        "name": config.get("name"),
        "title": config.get("title"),
        "description": config.get("description", ""),
        "repository": repository,
        "branch": branch,
        "version_identifier": version_identifier,
    }

    return pack_data


def create_or_update(endpoint, data, full_name):
    """
    Creates or updates Pack or Action or ConfigScheme object.
    Depends on the 'endpoint' argument.

    If the 'endpoint' is different than for Pack model,
    the 'pack_full_name' argument is mandatory.
    """

    model = endpoint.split("/")[-1][:-1]

    get_obj_url = endpoint + f"?full_name={full_name}"
    logger.info("Checking existence at %s", get_obj_url)
    obj_response = r.get(get_obj_url)

    if (obj_response.status_code // 100) != 2:
        raise click.ClickException(
            "Could not create or update. API status code: %s, data: %s"
            % (obj_response.status_code, obj_response.text)
        )

    count = obj_response.json()["count"]

    response = {}

    if count == 0:
        logger.info(f"Creating the {model}")
        logger.debug(f"Creating the %s at %s: %s", model, endpoint, data)
        response = r.post(endpoint, json=data)
    elif count == 1:
        obj_id = obj_response.json()["results"][0]["id"]
        obj_name = data["title"] if data.get("title") else data["name"]
        obj_url = endpoint + f"/{obj_id}"

        logger.info(f"Updating the {model}: {obj_name}")
        response = r.put(obj_url, json=data)
    else:
        logger.error(f"{obj_response.json()}")
        raise click.ClickException(
            "Fatal error. More than one object with these values. \
                                    Please contact with our support."
        )

    if (response.status_code // 100) != 2:
        raise click.ClickException(
            "Could not create or update. API status code: %s, data: %s"
            % (response.status_code, response.text)
        )

    return response.json()


def clone_repository(repository_url, path, branch=None):
    """
    Clone Git repository from "repository_url" into the "path"
    from the "branch" branch.
    """

    if branch is None:
        branch_info = "default branch"
    else:
        branch_info = f"branch {branch}"

    logger.info(
        "Cloning %s of repository from %s to %s", branch_info, repository_url, path
    )

    repo = git.Repo.clone_from(repository_url, path, branch=branch)
    active_branch_name = repo.active_branch.name
    commit_hash = repo.heads[active_branch_name].commit.hexsha
    logger.info("HEAD is at %s on branch %s", commit_hash, repo.active_branch)
    return commit_hash


def get_namespace(url):
    """
    Returns the namespace of the user based on API url.
    """
    parsed_url = urlparse(url)
    namespace = parsed_url.netloc.split(".")[0]

    if namespace == "api":
        return "njinn"

    return namespace


def check_api(api):
    """
    Checks if 'NJINN_URL' environment variable was overwritten via CLI argument
    or if it was set; returns it, otherwise displays prompt to provide value.

    Remove the trailing slash if found.
    """

    if not api:
        api = os.environ.get("NJINN_URL")
        if not api:
            api = input("API url: ")

    if api.endswith("/"):
        api = api[:-1]

    return api


def check_username(username):
    """
    Checks if 'NJINN_USER' environment variable was overwritten via CLI argument
    or if it was set; returns it, otherwise displays prompt to provide value.
    """

    if not username:
        username = os.environ.get("NJINN_USER")
        if not username:
            username = input("Username: ")

    return username


def check_password(password):
    """
    Checks if 'NJINN_PASS' environment variable was overwritten via CLI argument
    or if it was set; returns it, otherwise displays prompt to provide value.
    """

    if not password:
        password = os.environ.get("NJINN_PASS")
        if not password:
            password = getpass.getpass()

    return password


def zip_dir(source, target_temp_dir):
    logger.info("Creating zipfile in %s from %s", target_temp_dir, source)
    files = []
    git_path = os.path.normpath(os.path.join(source, ".git"))
    for dirpath, _, filenames in os.walk(source):
        # ignore git metadata
        if os.path.normpath(dirpath).startswith(git_path):
            continue
        files.extend([os.path.join(dirpath, filename) for filename in filenames])
    if not files:
        raise ValueError("No files found")
    zip_filename = f"{target_temp_dir}/bundle.zip"
    with ZipFile(zip_filename, mode="w") as zip_file:
        for repo_file in files:
            logger.info("Zipping %s", repo_file)
            zip_file.write(repo_file, arcname=os.path.relpath(repo_file, source))
    return zip_filename


def install(repository_url, branch, username, password, api, global_):
    os.environ["GIT_TERMINAL_PROMPT"] = "0"

    api = check_api(api)
    username = check_username(username)
    logger.info(f"Login user: {username} @ {api}")
    password = check_password(password)
    login(username, password, api)

    temp_dir = tempfile.TemporaryDirectory().name
    try:
        with tempfile.TemporaryDirectory() as target_temp_dir:
            commit_hash = clone_repository(repository_url, temp_dir, branch=branch)
            zip_filename = zip_dir(temp_dir, target_temp_dir)
            pack_conf_name = (
                "pack.yaml" if os.path.exists(f"{temp_dir}/pack.yaml") else "pack.yml"
            )
            pack_conf_path = os.path.join(temp_dir, pack_conf_name)

            if not os.path.exists(pack_conf_path):
                raise click.ClickException(
                    "No configuration file in the repository root."
                )

            try:
                with open(pack_conf_path, "r") as pack_conf:
                    logger.info("Loading the configuration file: %s" % pack_conf_name)
                    config = yaml.safe_load(pack_conf)

                    base_api = api + "/api/v1"
                    namespace = NJINN_NAMESPACE if global_ else get_namespace(base_api)
                    logger.info("Namespace used: %s", namespace)
                    pack_data = prepare_pack(
                        config, repository_url, branch, commit_hash
                    )
                    if global_:
                        pack_data["namespace"] = namespace
                    logger.info("Pack data %s", pack_data)
                    config_schemes_data = prepare_config_schemes(config)
                    actions_data = prepare_actions(config)

                    pack_full_name = ".".join([namespace, pack_data["name"]])
                    packs_url = base_api + "/packs"
                    pack = create_or_update(packs_url, pack_data, pack_full_name)

                    pack_id = pack["id"]
                    pack_url = pack["url"]

                    zip_basename = os.path.basename(zip_filename)
                    logger.info("Uploading %s", zip_basename)
                    bundle_url = f"{base_api}/packs/{pack_id}/bundle"
                    res = r.put(
                        bundle_url,
                        files={
                            "bundle": (zip_basename, open(zip_filename, "rb")),
                            "version_identifier": ("", commit_hash),
                        },
                    )
                    if not res.status_code == 200:
                        logger.warning("Response: %s - %s", res.status_code, res.text)
                        raise Exception(res.text)

                    if config_schemes_data:
                        config_schemes_url = base_api + "/config_schemes"
                        for config_scheme in config_schemes_data:
                            config_full_name = ".".join(
                                [pack_full_name, config_scheme["name"]]
                            )
                            config_scheme["pack"] = pack_url
                            create_or_update(
                                config_schemes_url, config_scheme, config_full_name
                            )

                    actions_url = base_api + "/actions"
                    for action in actions_data:
                        action_full_name = ".".join([pack_full_name, action["name"]])
                        action["pack_id"] = pack_id
                        create_or_update(actions_url, action, action_full_name)
            except OSError as e:
                raise click.ClickException("Could not open the file. %s" % str(e))
            except KeyError as e:
                raise click.ClickException(
                    "Invalid configured %s file. %s" % (pack_conf_name, e)
                )
    finally:
        # Borrowed from njinn/worker#57
        def on_rm_error(func, path, exc_info):
            os.chmod(path, stat.S_IWRITE)
            os.unlink(path)
        shutil.rmtree(temp_dir, onerror=on_rm_error)
        logger.info("Installation was successful")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pip install twine

import io
import os
import sys
from shutil import rmtree

from setuptools import setup, Command


NAME = "njinn"
DESCRIPTION = "Njinn CLI"
AUTHOR = "Njinn Technologies GmbH"
EMAIL = "contact@njinn.io"
REQUIRES_PYTHON = ">=3.7"
VERSION = "0.3.0"
REQUIRED = [
    "requests >=2.21, <2.22",
    "click >=7.0, <8.0",
    "gitpython >=2.1, <2.2",
    "pyyaml >=5.1, <5.2",
    "click-log >=0.3, <0.4",
    "click-spinner >=0.1, <0.2",
]
EXTRAS = {}


here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel (universal) distribution…")
        os.system("{0} setup.py sdist bdist_wheel".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        self.status("Pushing git tags…")
        os.system("git tag v{0}".format(about["__version__"]))
        os.system("git push --tags")

        sys.exit()


setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    packages=["njinn", "njinn.models", "njinn.resources"],
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    entry_points={"console_scripts": ["njinn=njinn.__main__:cli"],},
    include_package_data=True,
    license="Other/Proprietary License",
    classifiers=[
        "License :: Other/Proprietary License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: Implementation :: CPython",
        "Operating System :: OS Independent",
    ],
    cmdclass={"upload": UploadCommand,},
)



import logging

import click
import click_log

from .packs import install as pack_install

logger = logging.getLogger("njinn")
click_log.basic_config(logger)


@click.group()
def cli():
    pass


@cli.group()
def pack():
    pass


@pack.command()
@click.argument("repository_url")
@click_log.simple_verbosity_option(logger)
@click.option(
    "--branch",
    default="master",
    show_default=True,
    help="A branch from which the repository should be cloned.",
)
@click.option(
    "-u", "--username", "username", help="Username (Email Address) in Njinn API."
)
@click.option("-p", "--password", "password", help="Password for user in Njinn API.")
@click.option("-a", "--api", "api", help="Njinn API url.")
@click.option("--global", "global_", is_flag=True, help="Whether to install a pack globally. Works only with superuser permissions.")
def install(repository_url, branch, username, password, api, global_):
    logger.info("Installing pack from %s" % repository_url)
    pack_install(repository_url, branch, username, password, api, global_)


if __name__ == "__main__":
    cli()



from typing import List, TypeVar, Union

from ..base import *
from ..models import *
from .mixins import *

T = TypeVar("T")


class NjinnGroup(ParentMixin, SaveModelMixin, DeleteModelMixin, BaseModel):
    def __init__(
        self,
        id=None,
        name=None,
        permissions=None,
        group_members=None,
        created_at=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.id = id
        self.name = name
        self.permissions = permissions
        self.group_members = group_members
        self.created_at = created_at

    _read_only = ["id", "created_at"]

    _submodels = [GroupMember]

    def members(
        self, identifier=None, **kwargs
    ) -> Union[GroupMember, List[GroupMember]]:
        return self._get(GroupMember, identifier, **kwargs)



from ..base import *
from ..models import *
from .mixins import *


class TaskResource(GetResourceMixin, BaseResource):
    def __init__(self, parent: BaseResource):
        super().__init__("tasks", Task, parent.client)
        self.parent = parent

    def construct(self, response):
        return Task(response, api_resource=self)

Mapping.register(Task, TaskResource)


from ..base import *
from ..models import *
from .mixins import *


class TaskExecutionResource(
    GetResourceMixin, BaseResource,
):
    def __init__(self, parent: BaseResource):
        super().__init__("tasks", TaskExecution, parent.client)
        self.parent = parent

    def cancel(self, **kwargs) -> None:
        return self.operation("cancel").post(**kwargs)

    def log(self, **kwargs) -> str:
        return self.operation("log", result_class=str).get(**kwargs)


Mapping.register(TaskExecution, TaskExecutionResource)
